@@ -21,8 +21,8 @@ import (
 	"context"
 	"crypto/tls"
 	"fmt"
-	"github.com/IrineSistiana/mosdns/dispatcher/handler"
-	"github.com/IrineSistiana/mosdns/dispatcher/pkg/dnsutils"
+	"github.com/IrineSistiana/mosdns/v2/dispatcher/handler"
+	"github.com/IrineSistiana/mosdns/v2/dispatcher/pkg/dnsutils"
 	"github.com/miekg/dns"
 	"go.uber.org/zap"
 	"net"
@@ -98,7 +98,6 @@ func (s *Server) serveTCP(l net.Listener) error {
 
 				go func() {
 					qCtx := handler.NewContext(q, c.RemoteAddr())
-					qCtx.SetTCPClient(true)
 					s.handleQuery(tcpConnCtx, qCtx, &tcpResponseWriter{c: c})
 				}()
 			}
