@@ -18,6 +18,7 @@
 package upstream
 
 import (
+	"context"
 	"errors"
 	"fmt"
 	"github.com/IrineSistiana/mosdns/v2/dispatcher/pkg/pool"
@@ -26,19 +27,20 @@ import (
 	"io"
 	"net"
 	"sync"
-	"sync/atomic"
 	"time"
 )
 
 var (
-	errDialTimeout = errors.New("dial timeout")
-	errReadTimeout = errors.New("read timeout")
-	errIdCollision = errors.New("id collision")
+	errReadTimeout   = errors.New("read timeout")
+	errIdCollision   = errors.New("id collision")
+	errEOL           = errors.New("end of life")
+	errConnExhausted = errors.New("connection exhausted")
 )
 
-const (
-	defaultReadTimeout = time.Second * 5
-	defaultMaxConns    = 1
+var (
+	defaultReadTimeout     = time.Second * 5
+	defaultMaxConns        = 1
+	defaultMaxQueryPerConn = uint16(65535)
 )
 
 type Transport struct {
@@ -54,6 +56,11 @@ type Transport struct {
 	// Default is 1.
 	MaxConns int
 
+	// MaxConns controls the maximum queries that one connection
+	// handled. The connection will be closed if it reached the limit.
+	// Default is 65535.
+	MaxQueryPerConn uint16
+
 	// IdleTimeout controls the maximum idle time for each connection.
 	// If IdleTimeout <= 0, Transport will not reuse connections.
 	IdleTimeout time.Duration
@@ -84,6 +91,13 @@ func (t *Transport) maxConns() int {
 	return defaultMaxConns
 }
 
+func (t *Transport) maxQueryPerConn() uint16 {
+	if n := t.MaxQueryPerConn; n > 0 {
+		return n
+	}
+	return defaultMaxQueryPerConn
+}
+
 func (t *Transport) init() {
 	if logger := t.Logger; logger != nil {
 		t.logger = logger
@@ -96,12 +110,19 @@ func (t *Transport) init() {
 }
 
 type dialCall struct {
-	done chan struct{}
-	c    *clientConn // will be ready after done is closed.
-	err  error
+	waitingQId uint16
+	done       chan struct{}
+	c          *clientConn // will be ready after done is closed.
+	err        error
 }
 
 func (t *Transport) Exchange(q *dns.Msg) (r *dns.Msg, err error) {
+	ctx, cancel := context.WithTimeout(context.Background(), t.timeout())
+	defer cancel()
+	return t.ExchangeContext(ctx, q)
+}
+
+func (t *Transport) ExchangeContext(ctx context.Context, q *dns.Msg) (r *dns.Msg, err error) {
 	t.initOnce.Do(t.init)
 
 	if t.IdleTimeout <= 0 { // no keep alive
@@ -111,18 +132,18 @@ func (t *Transport) Exchange(q *dns.Msg) (r *dns.Msg, err error) {
 	start := time.Now()
 	retry := 0
 	for {
-		conn, reusedConn, err := t.getConn()
+		conn, reusedConn, qId, err := t.getConn(ctx)
 		if err != nil {
 			return nil, fmt.Errorf("no available connection, %w", err)
 		}
 
 		if !reusedConn {
-			return conn.exchange(q)
+			return conn.exchange(ctx, q, qId)
 		}
 
-		r, err = conn.exchange(q)
+		r, err = conn.exchange(ctx, q, qId)
 		if err != nil {
-			if time.Since(start) < time.Millisecond*200 && retry < 2 {
+			if !errors.Is(err, context.Canceled) && !errors.Is(err, context.DeadlineExceeded) && time.Since(start) < time.Millisecond*200 && retry <= 1 {
 				retry++
 				continue
 			}
@@ -157,36 +178,70 @@ func (t *Transport) removeConn(conn *clientConn) {
 	t.cm.Unlock()
 }
 
-func (t *Transport) getConn() (conn *clientConn, reusedConn bool, err error) {
+func (t *Transport) getConn(ctx context.Context) (conn *clientConn, reusedConn bool, qId uint16, err error) {
 	t.cm.Lock()
+	var availableConn *clientConn
 	for c := range t.conns {
+		if c.qId >= t.maxQueryPerConn() { // This connection has serverd too many queries.
+			// Note: the connection will close and clean up itself after its last query finished.
+			// We can't close it here. Some queries may still on that connection.
+			delete(t.conns, c)
+			continue
+		}
+		availableConn = c
+		break
+	}
+
+	if availableConn != nil && availableConn.onGoingQuery() == 0 { // An idle connection.
+		availableConn.qId++
+		qId = availableConn.qId
 		t.cm.Unlock()
-		return c, true, nil
+		return availableConn, true, qId, nil
 	}
 
-	// need a new connection
 	var dCall *dialCall
-	if len(t.dialingCalls) < t.maxConns() { // we can dial a new connection
-		dCall = t.startDial()
-	} else {
+	if len(t.conns)+len(t.dialingCalls) >= t.maxConns() {
+		// We have reached the limit and can't open a new connection.
+		if availableConn != nil { // We will reuse the connection.
+			availableConn.qId++
+			qId = availableConn.qId
+			t.cm.Unlock()
+			return availableConn, true, qId, nil
+		}
+
+		// No connection is available. Only dialingCalls.
+		// Wait a ongoing dial to complete.
 		for call := range t.dialingCalls {
+			if call.waitingQId >= t.maxQueryPerConn() { // To many waiting queries
+				continue
+			}
+			call.waitingQId++
+			qId = call.waitingQId
 			dCall = call
 			break
 		}
+	} else {
+		// No idle connection. Still can dial a new connection.
+		// Dial it now, to enlarge the capacity.
+		dCall = t.startDial()
+		qId = 0
 	}
 	t.cm.Unlock()
 
 	if dCall == nil {
-		panic("Transport getConn: dCall is nil")
+		return nil, false, 0, errConnExhausted
 	}
 
-	timer := pool.GetTimer(t.timeout())
-	defer pool.ReleaseTimer(timer)
 	select {
-	case <-timer.C:
-		return nil, false, errDialTimeout
+	case <-ctx.Done():
+		return nil, false, 0, ctx.Err()
 	case <-dCall.done:
-		return dCall.c, false, dCall.err
+		c := dCall.c
+		err := dCall.err
+		if err != nil {
+			return nil, false, 0, err
+		}
+		return c, false, qId, nil
 	}
 }
 
@@ -206,35 +261,36 @@ func (t *Transport) startDial() *dialCall {
 			t.cm.Unlock()
 			return
 		}
-		dConn := newDnsConn(t, c)
+		dConn := newClientConn(t, c)
 		dCall.c = dConn
 		close(dCall.done)
 		t.cm.Lock()
+		dConn.qId = dCall.waitingQId
 		t.conns[dConn] = struct{}{} // add dConn to conns
 		delete(t.dialingCalls, dCall)
 		t.cm.Unlock()
 
-		dConn.readLoop() // no needs to start a new goroutine
+		dConn.readLoop() // no need to start a new goroutine
 	}()
 	return dCall
 }
 
 type clientConn struct {
-	t *Transport
+	t   *Transport
+	qId uint16 // Managed and protected by t.
 
 	c net.Conn
 
-	qm    sync.RWMutex
-	queue map[uint16]chan *dns.Msg
+	qm      sync.RWMutex
+	queue   map[uint16]chan *dns.Msg
+	markEOL bool
 
 	cleanOnce sync.Once
 	closeChan chan struct{}
 	closeErr  error // will be ready after clientConn is closed
-
-	qId uint32 // atomic
 }
 
-func newDnsConn(t *Transport, c net.Conn) *clientConn {
+func newClientConn(t *Transport, c net.Conn) *clientConn {
 	return &clientConn{
 		t:         t,
 		c:         c,
@@ -243,14 +299,13 @@ func newDnsConn(t *Transport, c net.Conn) *clientConn {
 	}
 }
 
-func (c *clientConn) exchange(q *dns.Msg) (*dns.Msg, error) {
-	qId := uint16(atomic.AddUint32(&c.qId, 1))
-	qWithNewId := new(dns.Msg)
-	*qWithNewId = *q
-	qWithNewId.Id = qId
-
+func (c *clientConn) exchange(ctx context.Context, q *dns.Msg, qId uint16) (*dns.Msg, error) {
 	resChan := make(chan *dns.Msg, 1)
+
 	c.qm.Lock()
+	if qId >= c.t.maxQueryPerConn() {
+		c.markEOL = true
+	}
 	if _, ok := c.queue[qId]; ok {
 		c.qm.Unlock()
 		return nil, errIdCollision
@@ -261,9 +316,18 @@ func (c *clientConn) exchange(q *dns.Msg) (*dns.Msg, error) {
 	defer func() {
 		c.qm.Lock()
 		delete(c.queue, qId)
+		remain := len(c.queue)
+		markEOL := c.markEOL
 		c.qm.Unlock()
+
+		if markEOL && remain == 0 { // This is the last goroutine.
+			c.closeAndCleanup(errEOL)
+		}
 	}()
 
+	qWithNewId := new(dns.Msg)
+	*qWithNewId = *q
+	qWithNewId.Id = qId
 	c.c.SetWriteDeadline(time.Now().Add(generalWriteTimeout))
 	_, err := c.t.WriteFunc(c.c, qWithNewId)
 	if err != nil {
@@ -277,6 +341,8 @@ func (c *clientConn) exchange(q *dns.Msg) (*dns.Msg, error) {
 	select {
 	case <-timer.C:
 		return nil, errReadTimeout
+	case <-ctx.Done():
+		return nil, ctx.Err()
 	case r := <-resChan:
 		if r != nil {
 			r.Id = q.Id
@@ -324,3 +390,10 @@ func (c *clientConn) closeAndCleanup(err error) {
 		close(c.closeChan)
 	})
 }
+
+func (c *clientConn) onGoingQuery() int {
+	c.qm.RLock()
+	defer c.qm.RUnlock()
+
+	return len(c.queue)
+}
